#include "ResourceManager.H"
#include <algorithm>
#include "CommonUtility.H"
#include "SFXML.H"
#include "Direct3D.H"
#include "GraphicObject.H"

//////////////////////////////////////////////////////////////////////////////////////////////////
//	CFileResource
//////////////////////////////////////////////////////////////////////////////////////////////////

bool CFileResource::LoadResource()
{
	if(IsLoaded()) return true;
	
	ODS("Load Resource File [%s]\n",GetFileName());
	
	FILE *fp = fopen(GetFileName(),"rb");
	if(!fp) 
	{
		ODS("Failed - file open failed\n");
		return false;
	}

	fseek(fp,0,SEEK_END);
	DWORD dwEndPos = ftell(fp);
	fseek(fp,0,SEEK_SET);
	m_dwFileLength = dwEndPos - ftell(fp);

	if(m_dwFileLength == 0)
	{
		fclose(fp);
		ODS("Failed - file is empty\n");
		return true;
	}

	m_paBuffer = new BYTE[m_dwFileLength];

	fread(m_paBuffer,m_dwFileLength,1,fp);

	fclose(fp);
	
	ODS("Success\n");
	Loaded(true);
	return true;
}
bool CFileResource::DisposeResource()
{
	if(m_paBuffer) delete []m_paBuffer; m_paBuffer=NULL;
	m_dwFileLength=0;
	Loaded(false);
	return true;
}
//////////////////////////////////////////////////////////////////////////////////////////////////
//	CXMLResource
//////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLResource::LoadResource()
{
	if(IsLoaded()) return true;

	SFXMLParser	Parser;

	m_pXMLContent = Parser.ParseXMLFile(GetFileName());

	Loaded(true);

	return	true;
}
bool CXMLResource::DisposeResource()
{
	if(m_pXMLContent) delete m_pXMLContent;
	m_pXMLContent=NULL;

	Loaded(false);
	return true;
}
//////////////////////////////////////////////////////////////////////////////////////////////////
//	CTextureResource
//////////////////////////////////////////////////////////////////////////////////////////////////

bool CTextureResource::LoadResource()
{
	ODS("Load Texture Resource %s\n",GetFileName());

	g_D3D90.CreateTextureFromFile(&m_pSurface,GetFileName());

	Loaded(true);

	return	true;
}
bool CTextureResource::DisposeResource()
{
	g_D3D90.ReleaseTexture(GetTexture());
	m_pSurface = NULL;

	Loaded(false);

	return	true;
}
bool	CTextureResource::Destroy()
{ 
	DisposeResource();
	return	true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//	CResourceMgr
//////////////////////////////////////////////////////////////////////////////////////////////////

CResourceMgr* CResourceMgr::ms_Singleton;
CResource*	CResourceMgr::CreateNewResourceInstance(CResource::RESOURCE_TYPE Type,char *strFilename)
{
	CResource	*pNewRes = NULL;

	switch(Type)
	{
	case	CResource::RT_XML: pNewRes = new CXMLResource;	break;
	case	CResource::RT_FILE:	pNewRes = new CFileResource;	break;
	case	CResource::RT_TEXTURE:	pNewRes = new CTextureResource;	break;
	case	CResource::RT_MUSIC:	pNewRes = new CMusicResource;	break;
	case	CResource::RT_SOUND:	pNewRes = new CSoundResource;	break;
	default:
		{
			MessageBox(NULL,"Unknown Resource Type","CResourceMgr::LoadResource() Error",MB_OK);
			return NULL;
		}
	}
	
	pNewRes->SetFileName(strFilename);
	pNewRes->AddRefCount();

	return	pNewRes;
}
CResource*	CResourceMgr::LoadResource(CResource::RESOURCE_TYPE Type,char *strFilename,bool bUseCriticalSection)	//	Load res immediately
{
	assert(strFilename);

	if(bUseCriticalSection)	EnterCS();
	
	for(ResourceVectorItor Itor = m_vLoadedResource[Type].begin();Itor != m_vLoadedResource[Type].end(); Itor ++)
	{
		if(strcmp( (*Itor)->GetFileName(),strFilename) ==0 )
		{
			// already Loaded resource
			if(bUseCriticalSection)	LeaveCS();
			 (*Itor)->AddRefCount();
			return *Itor;
		}
	}

	CResource	*pNewRes = CreateNewResourceInstance(Type,strFilename);
	m_vLoadedResource[pNewRes->GetResourceType()].push_back(pNewRes);
	if(bUseCriticalSection)	LeaveCS();

	if(pNewRes->LoadResource())
	{
		return pNewRes;
	}


	return NULL;
}

bool	CResourceMgr::ReleaseResource(CResource *pRes)
{
	if(!pRes) return true;

	EnterCS();

	for(ResourceVectorItor Itor = m_vReservedResource[pRes->GetResourceType()].begin();Itor != m_vReservedResource[pRes->GetResourceType()].end(); Itor ++)
	{
		if((*Itor) == pRes)
		{
			pRes->MinusRefCount();
			if(pRes->GetRefCount() == 0)
			{
				m_vReservedResource[pRes->GetResourceType()].erase(Itor);
				delete pRes;
			}
			LeaveCS();
			return	true;
		}
	}

	for(Itor = m_vLoadedResource[pRes->GetResourceType()].begin();Itor != m_vLoadedResource[pRes->GetResourceType()].end(); Itor ++)
	{
		if((*Itor) == pRes)
		{
			pRes->MinusRefCount();
			if(pRes->GetRefCount() == 0)
			{
				m_vLoadedResource[pRes->GetResourceType()].erase(Itor);
				delete pRes;
			}
			LeaveCS();
			return	true;
		}
	}
	
	LeaveCS();

	return	true;
}
CResource*	CResourceMgr::ReserveResource(CResource::RESOURCE_TYPE Type,char *strFilename)	//	Load res in another thread
{
	assert(strFilename);
	EnterCS();

	for(ResourceVectorItor Itor = m_vReservedResource[Type].begin();Itor != m_vReservedResource[Type].end(); Itor ++)
	{
		if(strcmp( (*Itor)->GetFileName(),strFilename)==0 )
		{
			// already reserved resource
			LeaveCS();
			(*Itor)->AddRefCount();
			return (*Itor);
		}
	}

	for(Itor = m_vLoadedResource[Type].begin();Itor != m_vLoadedResource[Type].end(); Itor ++)
	{
		if(strcmp( (*Itor)->GetFileName(),strFilename) == 0 )
		{
			// already Loaded resource
			LeaveCS();
			(*Itor)->AddRefCount();
			return (*Itor);
		}
	}

	CResource	*pNewRes = CreateNewResourceInstance(Type,strFilename);

	m_vReservedResource[Type].push_back(pNewRes);

	ResumeThread(m_ThreadHandle);
	
	LeaveCS();

	return pNewRes;
}
DWORD WINAPI CResourceMgr::ThreadFunc(LPVOID lpv)
{
	g_ResourceMgr.LoadReservedResProcess();
	return 0;
}
bool	CResourceMgr::LoadReservedResProcess()
{
	ResourceVector	vLoadResource[CResource::RT_TOTAL_RESOURCE_TYPE];
	bool bNotExist ;
	ResourceVectorItor Itor,Itor2;
	CResource *pRes;
	bool	bLoaded;

	while(true)
	{
		bNotExist = true;

		EnterCS();

		if(m_bFinishThread) 
		{
			LeaveCS();
			break;
		}

		
		for(int i=0;i<CResource::RT_TOTAL_RESOURCE_TYPE;i++)
		{
			if(m_vReservedResource[i].size() == 0 ) continue;

			vLoadResource[i].reserve(m_vReservedResource[i].size());

			for(Itor = m_vReservedResource[i].begin(); Itor != m_vReservedResource[i].end(); Itor++)
			{
				pRes = *Itor;

				//	Check Already Loaded
				bLoaded = false;
				for(Itor2 = m_vLoadedResource[i].begin(); Itor2 != m_vLoadedResource[i].end(); Itor2++)
				{
					if((*Itor2)->IsSameResource(pRes))
					{
						bLoaded = true;
						break;
					}
				}

				if(!bLoaded) 
				{
					m_vLoadedResource[i].push_back(pRes);
					vLoadResource[i].push_back(pRes);
				}

				bNotExist = false;
			}

			m_vReservedResource[i].clear();
		}
		if(bNotExist) 
		{
			LeaveCS();
			SuspendThread(m_ThreadHandle);
		}
		else
			LeaveCS();

		for(i=0;i<CResource::RT_TOTAL_RESOURCE_TYPE;i++)
		{
			if(vLoadResource[i].size() == 0 ) continue;

			for(Itor = vLoadResource[i].begin(); Itor != vLoadResource[i].end(); Itor++)
			{
				pRes = *Itor;
				if(!pRes->LoadResource())
				{
					MyMessageBox("Resource Load Failed, Filename : %s",pRes->GetFileName());
				}
			}

			vLoadResource[i].clear();
		}
	}

	return true;
}
bool	CResourceMgr::Init()
{
	m_bFinishThread = false;

	InitializeCriticalSection(&m_CS);
	
	DWORD dwThreadID;
	m_ThreadHandle = CreateThread(NULL,0,ThreadFunc,0,CREATE_SUSPENDED,&dwThreadID);

	return true;
}
bool	CResourceMgr::Destroy()
{
	EnterCS();
	m_bFinishThread = true;
	LeaveCS();

	ResumeThread(m_ThreadHandle);

	DWORD dwExitCode;
	while(true)
	{
		GetExitCodeThread(m_ThreadHandle,&dwExitCode);
		if(dwExitCode != STILL_ACTIVE) break;

		Sleep(100);
	}

	CloseHandle(m_ThreadHandle);
	DeleteCriticalSection(&m_CS);


	for(int i=0;i<CResource::RT_TOTAL_RESOURCE_TYPE;i++)
	{
		for(ResourceVectorItor Itor = m_vReservedResource[i].begin();Itor != m_vReservedResource[i].end(); Itor ++)
		{
			delete *Itor;
		}

		for(Itor = m_vLoadedResource[i].begin();Itor != m_vLoadedResource[i].end(); Itor ++)
		{
			delete (*Itor);
		}

		m_vReservedResource[i].clear();
		m_vLoadedResource[i].clear();
	}

	return true;
}