//	Resource Manager
//	Classes in this source file, manage whole resource in this game engine

#ifndef _RESOURCE_MANAGER_
#define _RESOURCE_MANAGER_

#include "Common.H"
#include <mmsystem.h>
#include <vector>
#include <d3d9.H>
#include "Singleton.H"

using namespace std;


//	Base class of whole resource class
class	CResource
{
public:
	enum MEMORY_TYPE
	{
		MT_SYSTEM=0,
		MT_MANAGED,
		MT_DEVICE
	};

	enum RESOURCE_TYPE
	{
		RT_TEXTURE=0,
		RT_SOUND,
		RT_MUSIC,
		RT_XML,
		RT_SCRIPT,
		RT_FILE,
		RT_SPRITE,
		RT_TOTAL_RESOURCE_TYPE,
	};
private:
	char m_strFileName[MAX_RES_FILENAME_LENGTH+1];
	RESOURCE_TYPE	m_ResourceType;
	MEMORY_TYPE m_MemType;
	DWORD m_dwMemUse; // Occupied memory size in byte
	BYTE	m_byPriorityLevel; // Priority level ( 0(low priority) ~ 255(high priority) )
	DWORD	m_dwRefCount;
	DWORD	m_dwAccessCounter;
	DWORD	m_dwLastAccessTime;

	bool	m_bLoaded;

public:

	CResource() 
	{
		m_ResourceType = RT_FILE;
		m_MemType=MT_SYSTEM;
		m_dwMemUse=0;
		m_byPriorityLevel=0;
		m_dwAccessCounter=0;
		m_dwLastAccessTime=0;
		m_dwRefCount=0;
		m_bLoaded = false;
		m_strFileName[0]='\0';
	};

	virtual ~CResource() 
	{
		DisposeResource();
	};

	void	SetFileName(char *strFileName) 
	{
		assert(strFileName);
		assert(strlen(strFileName)<=MAX_RES_FILENAME_LENGTH);
		strcpy(m_strFileName,strFileName); 
	}
	char*	GetFileName()	{	return m_strFileName; }

	void SetPriorityLevel(BYTE byLevel) { m_byPriorityLevel = byLevel; }
	BYTE GetPriorityLevel()	{	return m_byPriorityLevel; }
	
	virtual bool SetMemoryType(MEMORY_TYPE MemType) { m_MemType = MemType; return true; }
	MEMORY_TYPE GetMemoryType() { return m_MemType; }

	RESOURCE_TYPE GetResourceType() { return m_ResourceType; }
	void	SetResourceType(RESOURCE_TYPE Type) {	m_ResourceType = Type;	}

	void UpdateLastAccessTime()	{	m_dwLastAccessTime = timeGetTime();	}
	DWORD GetLastAccessTime()	{	return m_dwLastAccessTime; }

	void AddAccessCounter()	{	m_dwAccessCounter++; }
	DWORD GetAccessCounter()	{	return m_dwAccessCounter; }

	void	Loaded(bool bLoaded) { m_bLoaded = bLoaded; }
	bool	IsLoaded()	{	return m_bLoaded; }

	DWORD	GetRefCount()	{	return m_dwRefCount;	}
	void	AddRefCount()	{	m_dwRefCount++;	}
	void	MinusRefCount()	{	if(m_dwRefCount>0) m_dwRefCount--;	}

	//	LoadResource()    !
	//	LoadResource  ִ ܼ ó ϵ Ұ!
	//	LoadResource , CResourceMgr  ReserveResource,LoadResource,ReleaseResource  ! ȣ !
	//	LoadResource  CResource  ӹ ٸ ҽ ε  !
	
	//	DisposeResource()    
	//	DisposeResource , CResourceMgr  ReserveResource,LoadResource,ReleaseResource  ! ȣ !

	virtual bool LoadResource() { m_bLoaded = true; return true; }	//	Ƽ  ó ɼ ִ ޼ҵ̹Ƿ !
	virtual bool DisposeResource() { m_bLoaded = false; return true; }	//	Ƽ  ó ɼ ִ ޼ҵ̹Ƿ !

	virtual bool IsSameResource(CResource *pRes)
	{
		if(!pRes) return false;
		if(pRes == this) return true;
		if(pRes->GetResourceType() != GetResourceType()) return false;

		if(strcmp(pRes->GetFileName(),GetFileName())==0) return true;

		return false;
	}
};	
class	CMusicResource	:	public	CResource
{
};
class	CSoundResource	:	public	CResource
{
};

class CFileResource : public CResource
{
	DWORD	m_dwFileLength;
	BYTE	*m_paBuffer;

public:

	CFileResource()
	{
		m_dwFileLength=0;
		m_paBuffer=NULL;
		SetResourceType(CResource::RT_FILE);
	}
	~CFileResource()
	{
		DisposeResource();
	}

	DWORD	GetBufferLength()	{	return	m_dwFileLength;	}
	BYTE*	GetBuffer()	{	return	m_paBuffer;	}

	bool LoadResource();
	bool DisposeResource();
};

class	SFXMLContent;
class CXMLResource : public CResource
{
	SFXMLContent	*m_pXMLContent;

public:

	CXMLResource()
	{
		m_pXMLContent = NULL;
		SetResourceType(CResource::RT_XML);
	}
	~CXMLResource()
	{
		DisposeResource();
	}
	SFXMLContent*	GetXMLContent()	{	return	m_pXMLContent;	}

	bool LoadResource();
	bool DisposeResource();
};

class	CTextureResource	:	public	CResource
{
	LPDIRECT3DTEXTURE9 m_pSurface;

public:

	CTextureResource()
	{
		m_pSurface = NULL;
	}
	~CTextureResource()	{	Destroy();	}

	LPDIRECT3DTEXTURE9	GetTexture()	{	return	m_pSurface;	}

	bool LoadResource();
	bool DisposeResource();
	
private:

	bool	Destroy();
};

//	Manager class of whole resource in this engine

typedef vector<CResource*> ResourceVector;
typedef ResourceVector::iterator ResourceVectorItor;

class	CResourceMgr	:	public Singleton<CResourceMgr>
{

	ResourceVector m_vLoadedResource[CResource::RT_TOTAL_RESOURCE_TYPE];	// ε ҽ
	ResourceVector m_vReservedResource[CResource::RT_TOTAL_RESOURCE_TYPE]; // εؾ  ҽ

	bool	m_bInCS;	//	Critical Section  ִ°?
	CRITICAL_SECTION m_CS;	//	Ƽ  ε  ũƼ 
	HANDLE m_ThreadHandle;
	bool m_bFinishThread;

public:

	CResourceMgr() { Init(); }
	~CResourceMgr() { Destroy(); }

	//	CResource*	LoadResource(CResource::RESOURCE_TYPE Type,char *strFilename,bool bUseCriticalSection = true);
	//	 ޼ҵ    
	//	 ޼ҵ bUseCriticalSection    ũƼ  Ѵ.
	//	,  ޼ҵ带 ȣϱ  ̹ ũƼ  ο ִٸ, bUseCriticalSection  ݵ false 
	//	 ޼ҵ带 ȣ ϱ  ũƼ  ۿ ִٸ, bUseCriticalSection  ݵ true  ؾ Ѵ.
	CResource*	LoadResource(CResource::RESOURCE_TYPE Type,char *strFilename,bool bUseCriticalSection = true);	//	Load res immediately
	CResource*	ReserveResource(CResource::RESOURCE_TYPE Type,char *strFilename);	//	Load res in another thread
	
	bool	ReleaseResource(CResource *pRes);	//	if it's refcount is zero, Remove from m_vLoadedResource,m_vReservedResource, and destroy it

	bool	NowInCS()	{	return	m_bInCS;	}
private:

	bool	Init();
	bool	Destroy();

	static DWORD WINAPI ThreadFunc(LPVOID lpv);

	bool	LoadReservedResProcess();
	CResource*	CreateNewResourceInstance(CResource::RESOURCE_TYPE Type,char *strFilename);

	void	EnterCS()
	{
		EnterCriticalSection(&m_CS);
		m_bInCS = true;
	}
	void	LeaveCS()
	{
		LeaveCriticalSection(&m_CS);
		m_bInCS = false;
	}
};

#define g_ResourceMgr CResourceMgr::GetSingleton()	//	for my convinience

#endif